using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace RobotCtrl
{
    public class Drive
    {
        public PositionInfo Position
        {
            get { lock (infoLock) { return info.Position; } }
        }

        public Track CurrentTrack { get { return track; } }

        public bool Power { set { ctrl.Power = value; } }
        public DriveInfo Info { get { return info; } }
        public bool Done { get { return track == null; } }

        public Drive(RunMode runMode)
        {
            if (!Config.IsWinCE) //THOMAS: Dieser Code ist obsolet!?! Nicht?
                runMode = RunMode.VIRTUAL;
            if (runMode == RunMode.REAL)
            {
                ctrl = new DriveCtrl_HW(Config.IODriveCtrl);
                left = new MotorCtrl_HW(Config.IOMotorCtrlLeft);
                right = new MotorCtrl_HW(Config.IOMotorCtrlRight);
            }
            else
            {
                ctrl = new DriveCtrl();
                left = new MotorCtrl();
                right = new MotorCtrl();
            }
            lock (drivesLock)
            {
                left.Acceleration = 10.0;
                right.Acceleration = 10.0;
            }
            runTracksThread = new Thread(RunTracks);
            runTracksThread.Priority = ThreadPriority.AboveNormal;
            runTracksThread.Start();
        }

        public void Reset()
        {
            ctrl.Reset();
        }

        public void Close()
        {
            ctrl.Reset();
            lock (drivesLock)
            {
                left.Close();
                right.Close();
            }
            runTracksThread.Abort();
        }

        public void Stop()
        {
            stop = true;
        }

        public void Halt()
        {
            halt = true;
        }

        public void WaitDone()
        {
            while (!Done) Thread.Sleep(Config.DefaultLoopWaitTime);
        }

        public void RunPause(double pauseTimeSeconds)
        {
            if (track == null)
                track = new TrackPause(pauseTimeSeconds);
        }

        public void RunLine(double length, double speed, double acceleration)
        {
            if (track == null)
                track = new TrackLine(length, speed, acceleration);
        }

        public void RunArcLeft(double radius, double angle, double speed, double acceleration)
        {
            if (track == null)
                track = new TrackArcLeft(radius, angle, speed, acceleration);
        }

        public void RunArcRight(double runRadius, double runAngle, double runSpeed, double runAcceleration)
        {
            if (track == null)
                track = new TrackArcRight(runRadius, runAngle, runSpeed, runAcceleration);
        }

        public void RunTurn(double runAngle, double runSpeed, double runAcceleration)
        {
            if (track == null)
                track = new TrackTurn(runAngle, runSpeed, runAcceleration);
        }

        public void RunATrack(Track track)
        {
            if (this.track != null)
                throw new InvalidOperationException("An other Track is allready running.");

            this.track = track;
        }

        void RunTracks()
        {
            double velocity = 0;
            double deltaTime;

            int ticks = Environment.TickCount;
            stop = false;
            halt = false;

            while (true)
            {
                // Mglichst schneller Process Control Loop
                // ----------------------------------------
                Thread.Sleep(Config.DefaultLoopWaitTime);

                // Stopp ?
                // -------
                if (stop)
                {
                    track = null;
                    stop = false;
                }

                // neuen Track starten ?
                // ---------------------
                if (track != oldTrack)
                {
                    if (track != null)
                    {
                        // Intitialisierung
                        // ----------------
                        lock (infoLock)
                        {
                            lock (drivesLock)
                            {
                                oldInfo.DistanceL = -left.Distance;
                                oldInfo.DistanceR = right.Distance;
                            }
                            info.Runtime = 0;
                        }
                        track.Start(info.Position);
                    }
                }
                oldTrack = track;

                // Aktuelle Prozessdaten erfassen
                // ------------------------------
                // Zeit
                int deltaTicks = Environment.TickCount - ticks;
                ticks += deltaTicks;
                deltaTime = deltaTicks / 1000.0;

                if (track != null)
                {
                    if ((track.Done) || ((halt && (velocity == 0))))
                    {
                        track = null;
                        halt = false;
                    }
                    else if (track.ResidualLength > 0)
                    {
                        // Neue Prozessparameter berechnen
                        // -------------------------------
                        if (halt)
                        {
                            velocity = Math.Max(0, velocity - deltaTime * track.acceleration);
                        }
                        else
                        {
                            // Beschleunigung
                            velocity += deltaTime * track.acceleration;

                            // Begrenzung auf Reisegeschwindigkeit
                            velocity = Math.Min(track.nominalSpeed, velocity);

                            // Verzgerung auf Zielposition
                            // Geschwindigkeit auf max. zulssige Bremsgeschwindigkeit limitieren
                            double ve;
                            double s = track.ResidualLength;
                            if (s >= 0)
                                ve = Math.Sqrt(2.0 * track.acceleration * s);
                            else
                                ve = 0;
                            if (double.IsNaN(ve))
                                ve = 0;
                            velocity = Math.Min(ve, velocity);
                            //System.Console.WriteLine(velocity);
                        }

                        // Geschwindigkeitsvorgabe fr linken/rechten Motor bestimmen 
                        double leftSpeed, rightSpeed;
                        track.IncrementalStep(deltaTime, velocity, out leftSpeed, out rightSpeed);

                        // Neue Prozessparameter aktivieren
                        // --------------------------------
                        left.Speed = leftSpeed;
                        right.Speed = rightSpeed;
                        left.Go();
                        right.Go();
                    }
                    else
                    {
                        track = null;
                    }
                }
                else
                {
                    // Idle-Zustand setzen
                    // -------------------
                    lock (drivesLock)
                    {
                        left.Speed = 0;
                        right.Speed = 0;
                        right.Go();
                        left.Go();
                    }
                }
                // Status speichern und Odometrie aktualisieren
                updateInfo(deltaTime);
            }
        }

        // Status speichern und Odometrie aktualisieren
        // ============================================
        void updateInfo(double timeInterval)
        {
            // Motor-Status
            info.DriveStatus = ctrl.Status;

            lock (infoLock)
            {
                lock (drivesLock)
                {
                    info.MotorStatusL = left.Status;
                    info.MotorStatusR = right.Status;

                    info.SpeedL = -left.Speed;
                    info.SpeedR = right.Speed;
                    info.DistanceL = -left.Distance;
                    info.DistanceR = right.Distance;
                }
                if (track != null)
                    info.Runtime = track.elapsedTime;
            }

            // Position und Richtung im Weltkoordinatensystem bestimmen 
            // --------------------------------------------------------
            double dL = info.DistanceL - oldInfo.DistanceL;
            double dR = info.DistanceR - oldInfo.DistanceR;

            double d;
            double x1, y1, phi1;
            double x2 = 0, y2 = 0, phi2 = 0;

            d = (dL + dR) / 2.0;
            x1 = info.Position.X;
            y1 = info.Position.Y;
            phi1 = info.Position.Angle / 180.0 * Math.PI;
            if (dL == dR)
            {
                x2 = x1 + d * Math.Cos(phi1);
                y2 = y1 + d * Math.Sin(phi1);
                phi2 = phi1;
            }
            else if (dL == -dR)
            {
                x2 = x1;
                y2 = y1;
                phi2 = phi1 + (dR / Config.AxleLength) * 2;
            }
            else
            {
                double r = Config.AxleLength * d / (dR - dL);
                double x0 = x1 + r * Math.Cos(phi1 + Math.PI / 2);
                double y0 = y1 + r * Math.Sin(phi1 + Math.PI / 2);
                double dphi = d / r;

                x2 = x0 + (x1 - x0) * Math.Cos(dphi) - (y1 - y0) * Math.Sin(dphi);
                y2 = y0 + (x1 - x0) * Math.Sin(dphi) + (y1 - y0) * Math.Cos(dphi);
                phi2 = phi1 + dphi;
            }
            phi2 = phi2 % (2 * Math.PI);
            lock (infoLock)
            {
                info.Position.X = x2;
                info.Position.Y = y2;
                info.Position.Angle = phi2 / Math.PI * 180;
                oldInfo.SetMembersFrom(info);
            }
        }

        // private members
        // ---------------
        Track track = null;
        Track oldTrack = null;
        DriveCtrl ctrl;
        MotorCtrl left;
        MotorCtrl right;
        private readonly DriveInfo info = new DriveInfo();
        private readonly DriveInfo oldInfo = new DriveInfo();
        bool stop;
        bool halt;
        Object infoLock = new object();
        Object trackLock = new object();
        Thread runTracksThread;
        Object drivesLock = new object();
    }
}
