﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;

namespace RobotCtrl
{
	public abstract class Track
	{
		public double nominalSpeed;
		public double acceleration;
		public double coveredLength;
		public double elapsedTime;

		protected bool reverse = false;
		protected double length = 0;
		protected double pauseTime = 0;
		protected double currentVelocity = 0;
		protected readonly PositionInfo startPosition = new PositionInfo();

        /// <summary>
        /// Fertig mit warten (TrackPause) oder fertig mit fahren (Alle anderen Track-Klassen).
        /// </summary>
		public virtual bool Done { get { return (elapsedTime > pauseTime) && (coveredLength > length); } }

        /// <summary>
        /// Noch zu fahrende länge. Wird wie auch die Eigenschaft Done von der Klasse Drive verwendet um zu prüfen,
        /// ob der Track am Ende ist (Es wird einfach anders geprüft und anders reagiert...)
        /// </summary>
		public virtual double ResidualLength { get { return length - coveredLength; } }

        /// <summary>
        /// Beginnt die Fahrt von der angegebenen Position.
        /// </summary>
        /// <param name="startPosition"></param>
		public virtual void Start(PositionInfo startPosition)
		{
			this.startPosition.SetMembersFrom(startPosition);
			this.coveredLength = 0;
			this.elapsedTime = 0;
		}

		protected void doStep(double timeInterval)
		{
			elapsedTime += timeInterval;
			coveredLength += timeInterval * currentVelocity;
		}

        /// <summary>
        /// Berechnung, die bei jedem Schritt ausgeführt wird. Anhand der neuen Gewschwindigkeit sollen die Geschwindigkeiten
        /// für den rechten und den linken Motor berechnet werden.
        /// ZU LETZT MUSS DIE METHODE doStep(double timeInterval) AUFGERUFEN WERDEN!
        /// </summary>
        /// <param name="timeInterval"></param>
        /// <param name="newVelocity"></param>
        /// <param name="leftSpeed"></param>
        /// <param name="rightSpeed"></param>
		public abstract void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed);
	}

	//----------------------------------------------------------------------------------------

	public class TrackPause : Track
	{
		public TrackPause(double pauseTimeSeconds)
		{
			this.pauseTime = pauseTimeSeconds;
		}

		public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
		{
			currentVelocity = newVelocity;
			leftSpeed = -newVelocity;
			rightSpeed = +newVelocity;
			doStep(timeInterval);
		}
	}

	//----------------------------------------------------------------------------------------

	public class TrackLine : Track
	{
		public TrackLine(double length, double speed, double acceleration)
		{
			reverse = (Math.Sign(length) ^ Math.Sign(speed)) != 0;
			this.length = Math.Abs(length);
			this.nominalSpeed = Math.Abs(speed);
			this.acceleration = acceleration;
		}

		public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
		{
			currentVelocity = newVelocity;
			if (reverse)
				newVelocity = -newVelocity;
			leftSpeed = -newVelocity;
			rightSpeed = +newVelocity;
			doStep(timeInterval);
		}
	}

	//----------------------------------------------------------------------------------------

	public class TrackArcLeft : Track
	{
		public TrackArcLeft(double radius, double angle, double speed, double acceleration)
		{
			reverse = (Math.Sign(angle) ^ Math.Sign(speed)) != 0;
			this.nominalSpeed = Math.Abs(speed);
			this.acceleration = acceleration;
			this.angle = Math.Abs(angle);
			this.radius = Math.Abs(radius);
			this.length = Math.Abs((radius + Config.AxleLength / 2.0) * 2.0 * Math.PI * angle / 360.0);
		}

		public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
		{
			currentVelocity = newVelocity;
			if (reverse)
				newVelocity = -newVelocity;
			leftSpeed = -newVelocity * (radius - Config.AxleLength / 2.0) / (radius + Config.AxleLength / 2.0);
			rightSpeed = newVelocity;
			doStep(timeInterval);
		}

		public double angle;
		public double radius;
	}

	//----------------------------------------------------------------------------------------

	public class TrackArcRight : Track
	{
		public TrackArcRight(double radius, double angle, double speed, double acceleration)
		{
			reverse = (Math.Sign(angle) ^ Math.Sign(speed)) != 0;
			this.nominalSpeed = Math.Abs(speed);
			this.acceleration = acceleration;
			this.angle = Math.Abs(angle);
			this.radius = Math.Abs(radius);
			this.length = Math.Abs((radius + Config.AxleLength / 2.0) * 2.0 * Math.PI * angle / 360.0);
		}

		public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
		{
			currentVelocity = newVelocity;
			if (reverse)
				newVelocity = -newVelocity;
			leftSpeed = -newVelocity;
			rightSpeed = newVelocity * (radius - Config.AxleLength / 2.0) / (radius + Config.AxleLength / 2.0); ;
			doStep(timeInterval);
		}

		public double angle;
		public double radius;
	}

	//----------------------------------------------------------------------------------------

	public class TrackTurn : Track
	{
		public TrackTurn(double angle, double speed, double acceleration)
		{
			reverse = (Math.Sign(angle) ^ Math.Sign(speed)) != 0;
			this.nominalSpeed = Math.Abs(speed);
			this.acceleration = acceleration;
			this.angle = Math.Abs(angle);
			this.length = Math.Abs(Config.AxleLength / 2.0 * 2.0 * Math.PI * angle / 360.0);
		}

		public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
		{
			currentVelocity = newVelocity;
			if (reverse)
				newVelocity = -newVelocity;
			rightSpeed = newVelocity;
			leftSpeed = newVelocity;
			doStep(timeInterval);
		}

		public double angle;
	}

	//----------------------------------------------------------------------------------------

}
