﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;

namespace RobotCtrl
{
    public sealed class TrackPylonsParams
    {
        /// <summary>
        /// Gibt an, wie nahe ein Objekt mindestens sein sollte, damit eine Kurve eingeleitet wird.
        /// Muss grösser sein als <see cref="NominalTurnRadius"/>.
        /// </summary>
        public double MinDistanceForTurn { get; set; }

        /// <summary>
        /// Gewünschter Idealradius einer Wende.
        /// Muss kleiner sein als <see cref="MinDistanceForTurn"/>.
        /// </summary>
        public double NominalTurnRadius { get; set; }

        /// <summary>
        /// Gewichtung für die Winkelkorrektur
        /// </summary>
        public double AngleCorrectionFactor { get; set; }

        /// <summary>
        /// Bestimmt, wie weit der Roboter nach einer Wende geradeaus fahren soll.
        /// </summary>
        public double LineDistanceAfterTurn { get; set; }

        /// <summary>
        /// Bestimmt, wie weit der Roboter nach dem Erkennen des Pylon noch gerade aus fährt, bevor eine Wende eingeleitet wird.
        /// </summary>
        public double LineDistanceBeforeTurn { get; set; }

        public TrackPylonsParams()
        {
            // Set Defaults
            this.MinDistanceForTurn = 1.1;
            this.NominalTurnRadius = 0.7;
            this.AngleCorrectionFactor = 10;
            this.LineDistanceAfterTurn = .4;
            this.LineDistanceBeforeTurn = .1;
        }
    }
    //public class TrackPylons : Track
    //{
    //    private readonly Radar radar;
    //    private readonly TrackPylonsParams trackParams;
    //    private Track runningTrack;
    //    private Track nextTrack;
    //    /// <summary>
    //    /// This is an endless Track
    //    /// </summary>
    //    public override bool Done { get { return false; } }

    //    public override double ResidualLength { get { return 100.0; /* never ending track */ } }
    //    public event EventHandler<PylonDetectionArgs> PylonDetectionChanged;

    //    public TrackPylons(double maxSpeed, double maxAcceleration, Radar radar, TrackPylonsParams trackParams)
    //    {
    //        if (radar == null)
    //            throw new ArgumentNullException("radar");
    //        if (trackParams == null)
    //            throw new ArgumentNullException("trackParams");

    //        this.radar = radar;
    //        this.trackParams = trackParams;

    //        reverse = false;
    //        this.nominalSpeed = Math.Abs(maxSpeed);
    //        this.acceleration = maxAcceleration;
    //        this.runningTrack = new TrackForcedLine(2, maxSpeed, maxAcceleration);
    //    }

    //    public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
    //    {
    //        currentVelocity = newVelocity;

    //        double distance = radar.Distance;

    //        //Throw object detection event
    //        if (PylonDetectionChanged != null)
    //        {
    //            if (distance > trackParams.MinDistanceForTurn)
    //            {
    //                //in case of no object is detected right now
    //                PylonDetectionArgs args = new PylonDetectionArgs(false);
    //                PylonDetectionChanged(this, args);
    //            }
    //            else
    //            {
    //                //in case there is an object detected right now
    //                PylonDetectionArgs args = new PylonDetectionArgs(true);
    //                PylonDetectionChanged(this, args);
    //            }
    //        }

    //        if (runningTrack != null && nextTrack != null &&
    //            runningTrack.Done)
    //        {
    //            runningTrack = nextTrack;
    //            nextTrack = null;
    //        }
    //        else if ((runningTrack is TrackArcLeft || runningTrack is TrackForcedLine) && !runningTrack.Done)
    //        {
    //            // ensure the arc/"first line" are allways finished
    //        }
    //        else if (distance > trackParams.MinDistanceForTurn)
    //        {
    //            if (runningTrack == null || !(runningTrack is TrackLine))
    //            {
    //                runningTrack = new TrackLine(10, newVelocity, acceleration);
    //            }
    //        }
    //        else
    //        {
    //            if (runningTrack == null || !(runningTrack is TrackArcLeft))
    //            {
    //                double turnAngle = 180 - (trackParams.NominalTurnRadius - distance) * trackParams.AngleCorrectionFactor;
    //                runningTrack = new TrackArcLeft((distance + trackParams.NominalTurnRadius) / 2, turnAngle, newVelocity, acceleration);

    //                // Vom Pylon wegfahren, damit nicht gleich mit dem nächsten
    //                // Halbkreis begonnen wird.
    //                nextTrack = new TrackForcedLine(trackParams.LineDistanceAfterTurn, newVelocity, acceleration);
    //            }
    //        }

    //        runningTrack.IncrementalStep(timeInterval, newVelocity, out leftSpeed, out rightSpeed);
    //        this.doStep(timeInterval);
    //    }

    //    /// <summary>
    //    /// Linie, die zwingend zu Ende gefahren werden muss (z.B. am 1. Pylon vorbei fahren oder nach einem Halbkreis)
    //    /// </summary>
    //    private class TrackForcedLine : TrackLine
    //    {
    //        public TrackForcedLine(double length, double speed, double acceleration)
    //            : base(length, speed, acceleration) { }
    //    }

    //    public class PylonDetectionArgs : System.EventArgs
    //    {

    //        private bool detected;

    //        public bool Detected
    //        {
    //            get { return detected; }
    //        }

    //        public PylonDetectionArgs(bool detected)
    //        {
    //            this.detected = detected;
    //        }
    //    }
    //}

    public class TrackPylons : Track
    {
        private readonly Radar radar;
        private readonly TrackPylonsParams trackParams;

        private readonly Queue<Track> tracks = new Queue<Track>();

        /// <summary>
        /// This is an endless Track
        /// </summary>
        public override bool Done { get { return false; } }

        public override double ResidualLength { get { return 100.0; /* never ending track */ } }
        public event EventHandler<PylonDetectionArgs> PylonDetectionChanged;

        public TrackPylons(double maxSpeed, double maxAcceleration, Radar radar, TrackPylonsParams trackParams)
        {
            if (radar == null)
                throw new ArgumentNullException("radar");
            if (trackParams == null)
                throw new ArgumentNullException("trackParams");

            this.radar = radar;
            this.trackParams = trackParams;

            reverse = false;
            this.nominalSpeed = Math.Abs(maxSpeed);
            this.acceleration = maxAcceleration;
            this.tracks.Enqueue(new TrackForcedLine(2, maxSpeed, maxAcceleration));
        }

        public override void IncrementalStep(double timeInterval, double newVelocity, out double leftSpeed, out double rightSpeed)
        {
            currentVelocity = newVelocity;

            double distance = radar.Distance;

            //Raise object detection event
            if (PylonDetectionChanged != null)
            {
                if (distance > trackParams.MinDistanceForTurn)
                {
                    //in case of no object is detected right now
                    PylonDetectionArgs args = new PylonDetectionArgs(false);
                    PylonDetectionChanged(this, args);
                }
                else
                {
                    //in case there is an object detected right now
                    PylonDetectionArgs args = new PylonDetectionArgs(true);
                    PylonDetectionChanged(this, args);
                }
            }

            Track currentTrack = tracks.Any() ? tracks.Peek() : null;
            if (currentTrack != null && currentTrack.Done)
            {
                tracks.Dequeue();
            }
            else if (currentTrack is TrackArcRight || currentTrack is TrackArcLeft || currentTrack is TrackForcedLine)
            {
                // ensure the arc/"first line" are allways finished
            }
            else if (distance <= trackParams.MinDistanceForTurn)
            {
                double turnAngle = 180 - (trackParams.NominalTurnRadius - distance) * trackParams.AngleCorrectionFactor;
                tracks.Clear();

                tracks.Enqueue(new TrackForcedLine(trackParams.LineDistanceBeforeTurn, newVelocity, acceleration));

                tracks.Enqueue(new TrackArcRight((distance + trackParams.NominalTurnRadius) / 2, turnAngle, newVelocity, acceleration));

                // Vom Pylon wegfahren, damit nicht gleich mit dem nächsten
                // Halbkreis begonnen wird.
                tracks.Enqueue(new TrackForcedLine(trackParams.LineDistanceAfterTurn, newVelocity, acceleration));

                currentTrack = tracks.Peek();
            }
            else if(!tracks.Any())
            {
                tracks.Enqueue(new TrackLine(10, newVelocity, acceleration));
                currentTrack = tracks.Peek();
            }

            currentTrack.IncrementalStep(timeInterval, newVelocity, out leftSpeed, out rightSpeed);
            this.doStep(timeInterval);
        }

        /// <summary>
        /// Linie, die zwingend zu Ende gefahren werden muss (z.B. am 1. Pylon vorbei fahren oder nach einem Halbkreis)
        /// </summary>
        private class TrackForcedLine : TrackLine
        {
            public TrackForcedLine(double length, double speed, double acceleration)
                : base(length, speed, acceleration) { }
        }

        public class PylonDetectionArgs : System.EventArgs
        {

            private bool detected;

            public bool Detected
            {
                get { return detected; }
            }

            public PylonDetectionArgs(bool detected)
            {
                this.detected = detected;
            }
        }

    }
}
