﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.Threading;

namespace Chart
{
    public partial class ChartView : UserControl
    {
        public ChartView()
        {
            InitializeComponent();

            this.Autoscale = false;

            values = new Queue<Double>();
            fillQueue(values, capacity);
        }

        private Queue<Double> values;

        private double scaleFactor;
        public double ScaleFactor
        {
            get { return scaleFactor; }
        }

        private bool autoscale;
        public bool Autoscale
        {
            get { return autoscale; }
            set { autoscale = value;
            if (value == false)
            {
                scaleFactor = 1;
            }
            }
        }

        private int capacity;
        public int Capacity
        {
            get { return capacity; }
            set { capacity = value; }
        }

        public void Enqueue(double value)
        {
            //lock values to not read/write at the same time
            lock (values)
            {
                if (values.Count > capacity)
                {
                    values.Dequeue();
                }
                values.Enqueue(value);
            }

            //rescale
            if (this.autoscale)
            {
                Rescale();
            }

            //refresh chart
            if (this.InvokeRequired)
            {
                this.Invoke(new Action(this.Refresh));
            }
            
        }

        private void Rescale()
        {
            double maxValue = values.Max();
            if (maxValue != 0)
            {
                scaleFactor = Math.Min(this.Height / maxValue, Int32.MaxValue);
            }
        }

        private void fillQueue(Queue<double> queue, int capacity)
        {
            for (int i = 0; i < capacity; i++)
            {
                queue.Enqueue(0);
            }
        }

        // This method draws the line graph.
        private void DrawLineGraph(Graphics g)
        {
            // Get the count into a local variable
            //   so that it's a little easier and the
            //   code is more readable.
            //int capacity = data.Count;

            // Calculate the distance between data points.
            int nDataSeparation = (this.Size.Width / capacity);

            // The panel height is used to calculate the
            //   y coordinate for drawing.
            int nPanelHeight = this.Size.Height;

            // We'll use this as an offset into the panel
            //   since we want some space from the left margin.
            int nXOffset = nDataSeparation / 2;

            

            // Create the two pens we'll use. We don't want
            //   to create each one for every loop iteration
            //   since that will be a performance slow down.
            Pen objVerticalPen = new Pen(Color.Gray);
            objVerticalPen.DashStyle = System.Drawing.Drawing2D.DashStyle.Dash;
            Pen penFrame = new Pen(Color.Gray, 2);
            Pen objLinePen = new Pen(Color.Magenta, 3);

            g.DrawLine(penFrame, 0, 0, 0, this.Size.Height);
            g.DrawLine(penFrame, 0, this.Size.Height, this.Size.Width, this.Size.Height);

            // Loop through the data.
            int i = -1;
            double value = 0;
            lock (values)
            {
                foreach (double nextValue in values)
                {
                    // Draw a vertical line so the user
                    //   gets a more understandable graph.
                    g.DrawLine(objVerticalPen, i * nDataSeparation + nXOffset,
                        nPanelHeight - Convert.ToInt32(value * scaleFactor),
                        i * nDataSeparation + nXOffset, nPanelHeight);

                    // If this isn't the last data item,
                    //   draw the connecting line to the next point.
                    //if (i < capacity - 1)
                    if (i >= 0)
                    {
                        g.DrawLine(objLinePen,
                            i * nDataSeparation + nXOffset,
                            nPanelHeight - Convert.ToInt32(value * scaleFactor),
                            (i + 1) * nDataSeparation + nXOffset,
                            nPanelHeight - Convert.ToInt32(nextValue * scaleFactor));
                    }
                    value = nextValue;
                    i++;
                }
            }
        }

        private void ChartView_Paint(object sender, PaintEventArgs e)
        {
            if (values != null)
            {
                DrawLineGraph(e.Graphics);
            }
        }



        /*
        // Call this method from the Form_Load method, passing your ZedGraphControl instance
        public void CreateChart(ZedGraphControl zgc)
        {
            // Get a reference to the GraphPane
            GraphPane myPane = zgc.GraphPane;

            // Set the titles and axis labels
            myPane.Title.Text = "Demonstration of Multi Y Graph";
            myPane.XAxis.Title.Text = "Time, s";
            myPane.YAxis.Title.Text = "Velocity, m/s";
            myPane.Y2Axis.Title.Text = "Acceleration, m/s2";

            // Make up some data points based on the Sine function
            PointPairList vList = new PointPairList();
            PointPairList aList = new PointPairList();
            PointPairList dList = new PointPairList();
            PointPairList eList = new PointPairList();

            // Fabricate some data values
            for (int i = 0; i < 30; i++)
            {
                double time = (double)i;
                double acceleration = 2.0;
                double velocity = acceleration * time;
                double distance = acceleration * time * time / 2.0;
                double energy = 100.0 * velocity * velocity / 2.0;
                aList.Add(time, acceleration);
                vList.Add(time, velocity);
                eList.Add(time, energy);
                dList.Add(time, distance);
            }

            // Generate a red curve with diamond symbols, and "Velocity" in the legend
            LineItem myCurve = myPane.AddCurve("Velocity",
               vList, Color.Red, SymbolType.Diamond);
            // Fill the symbols with white
            myCurve.Symbol.Fill = new Fill(Color.White);

            // Generate a blue curve with circle symbols, and "Acceleration" in the legend
            myCurve = myPane.AddCurve("Acceleration",
               aList, Color.Blue, SymbolType.Circle);
            // Fill the symbols with white
            myCurve.Symbol.Fill = new Fill(Color.White);
            // Associate this curve with the Y2 axis
            myCurve.IsY2Axis = true;

            // Generate a green curve with square symbols, and "Distance" in the legend
            myCurve = myPane.AddCurve("Distance",
               dList, Color.Green, SymbolType.Square);
            // Fill the symbols with white
            myCurve.Symbol.Fill = new Fill(Color.White);
            // Associate this curve with the second Y axis
            myCurve.YAxisIndex = 1;

            // Generate a Black curve with triangle symbols, and "Energy" in the legend
            myCurve = myPane.AddCurve("Energy",
               eList, Color.Black, SymbolType.Triangle);
            // Fill the symbols with white
            myCurve.Symbol.Fill = new Fill(Color.White);
            // Associate this curve with the Y2 axis
            myCurve.IsY2Axis = true;
            // Associate this curve with the second Y2 axis
            myCurve.YAxisIndex = 1;

            // Show the x axis grid
            myPane.XAxis.MajorGrid.IsVisible = true;

            // Make the Y axis scale red
            myPane.YAxis.Scale.FontSpec.FontColor = Color.Red;
            myPane.YAxis.Title.FontSpec.FontColor = Color.Red;
            // turn off the opposite tics so the Y tics don't show up on the Y2 axis
            myPane.YAxis.MajorTic.IsOpposite = false;
            myPane.YAxis.MinorTic.IsOpposite = false;
            // Don't display the Y zero line
            myPane.YAxis.MajorGrid.IsZeroLine = false;
            // Align the Y axis labels so they are flush to the axis
            myPane.YAxis.Scale.Align = AlignP.Inside;
            myPane.YAxis.Scale.Max = 100;

            // Enable the Y2 axis display
            myPane.Y2Axis.IsVisible = true;
            // Make the Y2 axis scale blue
            myPane.Y2Axis.Scale.FontSpec.FontColor = Color.Blue;
            myPane.Y2Axis.Title.FontSpec.FontColor = Color.Blue;
            // turn off the opposite tics so the Y2 tics don't show up on the Y axis
            myPane.Y2Axis.MajorTic.IsOpposite = false;
            myPane.Y2Axis.MinorTic.IsOpposite = false;
            // Display the Y2 axis grid lines
            myPane.Y2Axis.MajorGrid.IsVisible = true;
            // Align the Y2 axis labels so they are flush to the axis
            myPane.Y2Axis.Scale.Align = AlignP.Inside;
            myPane.Y2Axis.Scale.Min = 1.5;
            myPane.Y2Axis.Scale.Max = 3;

            // Create a second Y Axis, green
            YAxis yAxis3 = new YAxis("Distance, m");
            myPane.YAxisList.Add(yAxis3);
            yAxis3.Scale.FontSpec.FontColor = Color.Green;
            yAxis3.Title.FontSpec.FontColor = Color.Green;
            yAxis3.Color = Color.Green;
            // turn off the opposite tics so the Y2 tics don't show up on the Y axis
            yAxis3.MajorTic.IsInside = false;
            yAxis3.MinorTic.IsInside = false;
            yAxis3.MajorTic.IsOpposite = false;
            yAxis3.MinorTic.IsOpposite = false;
            // Align the Y2 axis labels so they are flush to the axis
            yAxis3.Scale.Align = AlignP.Inside;

            Y2Axis yAxis4 = new Y2Axis("Energy");
            yAxis4.IsVisible = true;
            myPane.Y2AxisList.Add(yAxis4);
            // turn off the opposite tics so the Y2 tics don't show up on the Y axis
            yAxis4.MajorTic.IsInside = false;
            yAxis4.MinorTic.IsInside = false;
            yAxis4.MajorTic.IsOpposite = false;
            yAxis4.MinorTic.IsOpposite = false;
            // Align the Y2 axis labels so they are flush to the axis
            yAxis4.Scale.Align = AlignP.Inside;
            yAxis4.Type = AxisType.Log;
            yAxis4.Scale.Min = 100;

            // Fill the axis background with a gradient
            myPane.Chart.Fill = new Fill(Color.White, Color.LightGoldenrodYellow, 45.0f);

            zgc.AxisChange();
        }*/
    }
}
